# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Controls are widgets that are manipulated and return a value. They may
    update their function on move and release.
"""
import logging
import math

import pygame
from custom_inherit import doc_inherit

from pi_touch_gui._widget_bases import IControlWidget

LOG = logging.getLogger(__name__)


class Dial(IControlWidget):
    """ Circular input control dial.
    """

    @doc_inherit(IControlWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size, **kwargs):
        """ Initialize the dial.
        """
        # Dials don't use label, so block it
        if kwargs.get('label'):
            raise ValueError("Dials don't use their value as the label.")

        self.size = size
        diameter = min(size[0], size[1])
        radius = diameter >> 1
        kw_font_size = kwargs.pop('font_size', None)
        font_size = kw_font_size or (radius >> 1)

        # The parent init sets convenience properties like self.x, etc
        super(Dial, self).__init__(position, (diameter, diameter),
                                   font_size=font_size,
                                   **kwargs)
        self._radius = radius
        self._cx, self._cy = (self._x + self._w // 2), (self._y + self._h // 2)

    def touch_inside(self, touch):
        """ The interior of the dial is determined by a circle.
        """
        x, y = touch.position
        distance = math.hypot(x - self._cx, y - self._cy)
        if distance <= self._radius:
            return True

        return False

    def _on_move(self, touch):
        # If too many touches, we bail
        if len(self._touches) > 1:
            return

        dx = touch._x - self._cx
        dy = touch._y - self._cy

        # Value is from 0.0 to 1.0 (percent of range)
        self._value = (math.atan2(dy, dx) % math.tau) / math.tau
        self._adjusted_value = self.map_value_to_adjusted_value(self._value)

        super(Dial, self)._on_move(touch)

    def render(self, surface):
        wedge_rect = (self._x, self._y, self._w, self._h)
        handle_pos = (
            int(self._cx + (self._radius * math.cos(self._value * math.tau))),
            int(self._cy + (self._radius * math.sin(self._value * math.tau))))

        center = (self._cx, self._cy)
        # Draw the filled overall dial
        pygame.draw.circle(surface, self.color1, center, self._radius, 0)
        # Draw the active pie-wedge
        pygame.draw.arc(surface, self.color2, wedge_rect,
                        -self._value * math.tau, 0.0,
                        int(self._radius * 0.5))
        # Draw the handle line
        pygame.draw.line(surface, self.label_color1, center, handle_pos,
                         3)
        # Draw the handle circle
        pygame.draw.circle(surface, self.label_color1, handle_pos,
                           int(self._radius * 0.1), 0)

        self.render_centered_text(surface, self.value_str(),
                                  self.label_color1, self.color1)

        if self._focus:
            self.render_focus(surface)


class Slider(IControlWidget):
    """ Variable value input linear slider
    """

    @doc_inherit(IControlWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size, **kwargs):
        """ Initialize the slider.

        color1 is the slider background fill
        color2 is the slider outline and active value fill
        label_color1 is the slider line (and label)
        """
        # Sliders don't use label, so block it
        if kwargs.get('label'):
            raise ValueError("Dials don't use their value as the label.")

        super(Slider, self).__init__(position, size, **kwargs)

    def _on_move(self, touch):
        # If too many touches, bail
        if len(self._touches) > 1:
            return

        x, y = touch.position

        # Compute X/Y relative to the slider
        x -= self._x
        y -= self._y

        # Check our ratios...
        if self._w > self._h:
            # Horizontal Slider
            if 0 <= x <= self._w:
                self._value = float(x) / float(self._w)

        elif self._h > self._w:
            # Vertical Slider
            if 0 <= y <= self._h:
                self._value = (self._h - float(y)) / float(self._h)

        self.adjusted_value = self.map_value_to_adjusted_value(self._value)

        super(Slider, self)._on_move(touch)

    def render(self, surface):
        pygame.draw.rect(surface, self.color1, (self.position, self.size), 0)

        if self._w > self._h:
            # Horizontal slider
            value_w = int(self._w * self._value)
            value_x = self._x + value_w
            pygame.draw.rect(surface, self.color2,
                             ((self._x, self._y), (value_w, self._h)))
            pygame.draw.line(surface, self.label_color1,
                             (value_x, self._y), (value_x, self._y + self._h),
                             3)

        elif self._h > self._w:
            # Vertical slider
            value_h = int(self._h * self._value)
            value_y = self._y + (self._h - value_h)
            pygame.draw.rect(surface, self.color2,
                             ((self._x, value_y), (self._w, value_h)))
            pygame.draw.line(surface, self.label_color1,
                             (self._x, value_y), (self._x + self._w, value_y),
                             3)

        pygame.draw.rect(surface, self.color2, (self.position, self.size),
                         3)

        self.render_centered_text(surface, self.value_str(), self.label_color1)

        if self._focus:
            self.render_focus(surface)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Graphical User Interface objects

    A `Page` is a set of widgets that define the interaction and display
    surface of the GUI

    The `GUI` manages the hardware and runs the Pages
"""
import importlib
import json
import logging
import math
from pathlib import Path
from typing import List, Optional

import pygame

from pi_touch_gui._utilities import backlog_error
from pi_touch_gui._widget_bases import IWidget, IWidgetInterface
from pi_touch_gui.colors import BLACK
from pi_touch_gui.multitouch.raspberrypi_ts import TS_RELEASE

LOG = logging.getLogger(__name__)


class Page(object):
    WALL = 'wall'
    GO_WALL = -1
    GO_UP = 0
    GO_DOWN = 1
    GO_LEFT = 2
    GO_RIGHT = 3

    GO_NEIGHBORLESS = [-1, 1, -1, 1]
    GO_NAMES = ["Up", "Down", "Left", "Right"]

    LCARS = 'lcars'

    def __init__(self, name, widgets=None, background=None, function=None):
        """ A collection of widgets that make up an interaction page.

        Parameters
        ----------
        name : String
            Name of the page for internal and callback reference
        widgets : List[IWidget]
        background : str
            Background image resource filepath for this Page
        function : Callable[[Page], Optional[Type['Page']]]
            Function to call if none of the widgets consume
            an event.
        """
        # If no widgets added on init, they can be added with `add_widgets()`
        # later.
        self.name = name
        self.widgets = widgets or []

        if background == Page.LCARS:
            background = str(Path(Path(__file__).parent,
                                       "assets/lcars_screen.png"))
        self.background = background
        self._bg_image = None

        self.function = function

        self._focus_widget = None
        if widgets is None:
            self.widgets = []
            self._prev_focus_widget = None
        else:
            self.widgets = widgets
            self._prev_focus_widget = widgets[0]

        self._last_adjustment = pygame.time.get_ticks()
        self._adjustment_rate = 100

    def set_background(self, filepath):
        self.background = filepath
        self._bg_image = pygame.image.load(filepath).convert()

    def add_widgets(self, widgets):
        """ Add widgets to the Page.

        Parameters
        ----------
        widgets : List[IWidget]
        """
        self.widgets += widgets
        # Set or reset the base focus widget to the top
        self._prev_focus_widget = self.widgets[0]

    def find_widget(self, widget_or_id):
        """ Given a widget ID, find that widget.

        Parameters
        ----------
        widget_or_id : Union[IWidgetInterface, String]
            If a widget is passed in, return it directly. Otherwise search
            for the id and return the widget.

        Returns
        -------
        IWidgetInterface
        """
        # Deal with the WALL case, where we _know_ there isn't a widget
        if widget_or_id == Page.WALL:
            return None
        if not isinstance(widget_or_id, str):
            return widget_or_id

        return self._scan_for_widget_by_id(self.widgets, widget_or_id)

    def _scan_for_widget_by_id(self, widgets, widget_id):
        """ Recursively scan for a widget ID, returning the widget.

        Parameters
        ----------
        widgets : List[IWidgetInterface]
            A list of widgets, some of which may container lists of widgets
        widget_id : String
            The widget ID to search for

        Returns
        -------
        IWidgetInterface
        """
        for widget in widgets:
            # See if this is a widget container...
            found_widget = None
            sub_widgets = widget.sub_widgets
            if sub_widgets:
                # Assume the widget is a widget container and assert if not
                found_widget = self._scan_for_widget_by_id(sub_widgets,
                                                           widget_id)
            else:
                if widget.id == widget_id:
                    found_widget = widget
            if found_widget is not None:
                return found_widget
        return None

    def find_next_widget(self, widget_or_id, direction):
        """ Given a widget ID, find the widget AFTER that widget.

        Parameters
        ----------
        widget_or_id : Union[IWidgetInterface, String]
        direction : Integer
            Direction of scan, sign(direction)

        Returns
        -------
        IWidgetInterface
        """
        # Deal with the WALL case, where we _know_ there isn't a widget
        if widget_or_id == Page.WALL:
            return None
        try:
            widget_or_id = widget_or_id.id
        except AttributeError:
            pass

        return self._scan_for_next_widget_by_id(self.widgets,
                                                widget_or_id,
                                                math.copysign(1, direction))

    def _scan_for_next_widget_by_id(self, widgets, widget_id, step):
        """ Recursively scan for a widget ID, returning the widget AFTER the
        widget found.

        Parameters
        ----------
        widgets : List[IWidgetInterface]
            A list of widgets, some of which may container lists of widgets
        widget_id : String
            The widget ID to search for
        step : Integer
            The step (+1, -1) we are scanning next for.

        Returns
        -------
        IWidgetInterface
        """
        step = int(step)
        for idx, widget in enumerate(widgets):
            # See if this is a widget container...
            found_widget = None
            sub_widgets = widget.sub_widgets
            if sub_widgets:
                # Assume the widget is a widget container and assert if not
                found_widget = self._scan_for_widget_by_id(sub_widgets,
                                                           widget_id)
            else:
                if widget.id == widget_id:
                    try:
                        found_widget = widgets[idx + step]
                    except Exception as e:
                        LOG.error(f"Unmanaged Exception {type(e)}: {e}")
                        return None
            if found_widget is not None:
                return found_widget
        return None

    def touch_handler(self, event, touch) -> Optional['Page']:
        """ Update all widgets with a touch event.

            The first widget to consume the touch wins, and downstream
            widgets won't be tested.

            If no widget consumes the event, we call `function(self) -> Page`

        Parameters
        ----------
        event : int
            Event code defined for the interface; e.g. _widget_bases.py imports
            TS_PRESS, TS_RELEASE, and TS_MOVE
        touch : Touch
            The Touch object for the event.

        Returns
        -------
        Optional[Page]
            If the widget function names a new page, it is returned here
        """
        for widget in self.widgets:
            consumed, new_page = widget.event(event, touch)
            if consumed:
                return new_page

        if event == TS_RELEASE and self.function:
            return self.function(self, False)
        return None

    def render(self, surface: pygame.Surface):
        """Redraw all widgets to screen.

        Call `render()` for all widgets in the Page, once per frame.

        Parameters
        ----------
        surface : pygame.Surface
            Surface (screen) to draw to
        """
        # Lazy loading to avoid pygame initialization issues
        if self._bg_image is None and self.background is not None:
            self._bg_image = pygame.image.load(self.background).convert()

        if self._bg_image is not None:
            surface.blit(self._bg_image, (0, 0))
        else:
            surface.fill(BLACK)
        for widget in self.widgets:
            try:
                widget.render(surface)
            except Exception as e:
                backlog_error(e, f"Problem rendering widget {widget}")
                raise

    def focus_widget(self, enable):
        """ Wake up the focus system for gesture control.

        Will always restart on the last widget that had focus.

        Parameters
        ----------
        enable : Bool
            If True, then focus; else de-focus
        """
        if self.is_focused() == enable:
            return

        prev_widget = self._prev_focus_widget
        self._focus_widget = self._prev_focus_widget
        if enable:
            if not prev_widget.can_focus():
                # This will walk the down chain until it finds a widget
                # that can take focus.  Skips starting labels and stuff.
                self.shift_widget_focus(Page.GO_DOWN)
            else:
                prev_widget.focus(True)
        else:
            self._focus_widget = None
            prev_widget.focus(False)

    def is_focused(self):
        """ Determine if we are in widget focus mode.

        Returns
        -------
        Bool
        """
        return self._focus_widget is not None

    def shift_widget_focus(self, direction):
        """ Shift focus one way or the other, wrapping at the boundaries.

        Parameters
        ----------
        direction : int
            Direction of motion; Page.WALL (no motion),
        """
        if direction == Page.GO_WALL:
            return
        if not self.is_focused():
            return

        start_widget = test_widget = self._focus_widget
        while True:
            next_widget = test_widget.neighbors[direction]

            # We have an IWidgetInterface, ID string, or 'wall'
            if next_widget == Page.WALL:
                # Hit a wall
                return
            elif next_widget is None:
                # No neighbor specified, so use a generic direction
                next_widget = self.find_next_widget(
                    test_widget,
                    Page.GO_NEIGHBORLESS[direction])
            else:
                # Neighbor as a widget ID or index
                next_widget = self.find_widget(next_widget)

            # In case we instantiated a neighbor, store it
            test_widget.neighbors[direction] = next_widget

            if next_widget == Page.GO_WALL:
                return
            elif next_widget == start_widget:
                return
            elif next_widget is None:
                return

            # Take the tentative next widget and make it current
            test_widget = next_widget
            if test_widget.can_focus():
                break

        try:
            if self._focus_widget is not None:
                self._focus_widget.focus(False)
            next_widget.focus(True)
        except Exception as e:
            LOG.error(f"Unmanaged Exception {type(e)}: {e}")
        self._focus_widget = next_widget
        self._prev_focus_widget = next_widget

    def adjust_widget(self, direction):
        """ Send an up/down value adjustment to the control widget.
        """
        time = pygame.time.get_ticks()
        if time - self._last_adjustment < self._adjustment_rate:
            return

        if self._focus_widget._adjust is not None:
            self._focus_widget._adjust(direction)
            self._last_adjustment = time

    def press_widget(self):
        """ Activate the function of the widget.
        """
        if self._focus_widget.press is not None:
            self._focus_widget.press()

    def release_widget(self):
        """ Activate the function of the widget.
        """
        if self._focus_widget.release is not None:
            return self._focus_widget.release()
        return None

    def iter_widgets(self):
        """ Iterate over the widgets in the page.

        Returns
        -------
        iter(IWidget)
        """
        return iter(self.widgets)

    def serialize(self, readable=False):
        """ Serialize this page and return it as a JSON string, suitable
        to be deserialized()
        """
        indent = 4 if readable else None
        return json.dumps(self._do_serialize(self),
                          indent=indent,
                          sort_keys=True)

    def _do_serialize(self, entity):
        if entity is None:
            return None
        if type(entity) in [str, int, float, complex, bool]:
            return entity
        elif type(entity) in [list, tuple, set]:
            # JSON doesn't distinguish list-equivalent types
            return [self._do_serialize(v) for v in entity]
        elif type(entity) is dict:
            return {k: self._do_serialize(v) for k, v in entity.items()}
        elif callable(entity):
            pause = True
            path = '.'.join([entity.__module__,
                             entity.__qualname__]).split('.')
            context = '.'.join(path[:-1])
            name = path[-1]

            if '<locals>' in context:
                raise ValueError(f'Can not serialize local function '
                                 f'{context}.{name}')
            if '<lambda>' in context:
                raise ValueError(f'Can not serialize lambda function '
                                 f'{context}.{name}')
            elif 'builtins' in context:
                raise ValueError(f'Can not serialize builtin function '
                                 f'{context}.{name}')
            return {'__Function': {'context': context,
                                   'name': name}}
        elif issubclass(type(entity), Page):
            return self._do_serialize_object(entity)
        elif issubclass(type(entity), IWidgetInterface):
            return self._do_serialize_object(entity)
        else:
            raise TypeError(f"Unsupported type {type(entity)!r}")

    def _do_serialize_object(self, entity):
        ent_type = type(entity)
        kwargs = {k: self._do_serialize(v) for k, v in vars(entity).items()
                  if not k.startswith('_')}
        return {'__Object': {'module': ent_type.__module__,
                             'class': ent_type.__name__,
                             'kwargs': kwargs}}

    @classmethod
    def deserialize(cls, serial):
        """ Deserialize the data and return the Page it started as.
        """
        data = json.loads(serial)
        return Page._do_deserialize(data)
        pass

    @classmethod
    def _do_deserialize(cls, entity):
        if entity is None:
            return None
        if type(entity) in [str, int, float, complex, bool]:
            return entity
        if type(entity) is list:
            return [Page._do_deserialize(v) for v in entity]
        elif type(entity) is dict:
            content = {}
            for key in entity.keys():
                if key == '__Function':
                    value = entity[key]
                    context = value['context']
                    func_name = value['name']
                    return Page._get_context(context, func_name)
                elif key == '__Object':
                    value = entity[key]
                    module = value['module']
                    class_name = value['class']
                    kwargs = Page._do_deserialize(value['kwargs'])
                    c = Page._get_context(module, class_name)
                    return c(**kwargs)
                else:
                    content.update({key: Page._do_deserialize(entity[key])})
            return content
        else:
            raise TypeError(f"Unsupported type {type(entity)!r}")

    @classmethod
    def _get_context(cls, module, item):
        path = module.split('.')
        imp = path[0]
        try:
            m = importlib.import_module(imp)
        except Exception as e:
            raise ImportError(f"Unable to import {imp!r}: {e}")
        for part in path[1:]:
            m = getattr(m, part)
        return getattr(m, item)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Buttons are widgets that are pressed. They typically update their function
    on release, and the contents can be as simple as a label and as complex
    as a multivalued graph.
"""
import logging
import uuid
from typing import Tuple, Optional

import pygame

from custom_inherit import doc_inherit

from pi_touch_gui._widget_bases import (IButtonWidget, IWidgetInterface)

LOG = logging.getLogger(__name__)


class Button(IButtonWidget):
    """ A simple rectangular button that calls its function on release.

    It is an empty instantiation of IButtonWidget.
    """
    # Button style
    RECTANGLE = 0
    LEFT_ROUND = 1
    RIGHT_ROUND = 2
    OBROUND = 3

    # Button behavior
    CLICK = 0
    HOLD = 1
    TOGGLE = 2

    @doc_inherit(IButtonWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size,
                 style=None,
                 behavior=None,
                 **kwargs):
        """ Initialize the Button.

        Parameters
        ----------
        style : Int
            The style in which the button is rendered (Button.RECTANGLE is the
            default, but you can also specify LEFT_ROUND, RIGHT_ROUND, or both
            with OBROUND)
        behavior : Int
            The behavior of the button (Button.CLICK is the default, activating
            on release, but you can specify HOLD which activates on press
            and stops on release, or TOGGLE which turns on and off on each
            release).
        """
        super(Button, self).__init__(position, size, **kwargs)
        self.style = style or Button.RECTANGLE
        self.behavior = behavior or Button.CLICK
        self._active = False

    @doc_inherit(IButtonWidget.render, style='numpy_with_merge')
    def render(self, surface):
        """ Render the button.

        Complicated by the need for one or two rounded ends.  If you play
        your parameters right, you get a circle!
        """
        color, label_color = self.state_colors(self._active)

        radius = int(self._h / 2)
        inset_x = radius if self.style & self.LEFT_ROUND else 0
        inset_w = radius if self.style & self.RIGHT_ROUND else 0
        inset_w = inset_w + inset_x
        subrect = ((self._x + inset_x, self._y),
                   (self._w - inset_w, self._h))
        pygame.draw.rect(surface, color, subrect, 0)
        if self.style & self.LEFT_ROUND:
            pygame.draw.circle(surface, color,
                               (self._x + radius, self._y + radius),
                               radius)
        if self.style & self.RIGHT_ROUND:
            pygame.draw.circle(surface, color,
                               (self._x + self._w - radius, self._y + radius),
                               radius)

        self.render_centered_text(surface, self.label, label_color, color)

        if self._focus:
            self.render_focus(surface)

    def set(self):
        if self.behavior == Button.TOGGLE:
            self._active = True
            if self.function:
                return self.function(self, self._active)

    def clear(self):
        if self.behavior == Button.TOGGLE:
            self._active = False
            if self.function:
                return self.function(self, self._active)

    def _on_press(self, touch) -> None:
        """ Hold behavior calls their function on the press event.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for buttons
        """
        self.press()

    def press(self):
        if self.behavior == self.HOLD:
            self._active = True
            if self.function:
                self.function(self, self._active)
        super(Button, self).press()

    def _on_release(self, touch):
        return self.release()

    def release(self):
        if self.behavior == self.TOGGLE:
            self._active = not self._active
        else:
            self._active = False
        return super(Button, self).release()


class ButtonGroup(IWidgetInterface):
    """ A group of ToggleButton where zero or one may be active.

    When a button is made active, all the remaining are
    forced to be inactive.

    It is possible for none of the buttons to be active, if you
    de-activate the active button.

    Starts with the first button in the list active.
    """

    @doc_inherit(IWidgetInterface.__init__, style='numpy_with_merge')
    def __init__(self, toggle_buttons, widget_id=None):
        """ Initialize the button group, and hook the button functions.

        Parameters
        ----------
        toggle_buttons : List[ToggleButton]
            An ordered list of ToggleButton objects, where the first button
            is active on startup.
        widget_id : str
            Optional ID for this group, if not given, gets a random uuid.
        """
        self._label = 'group'
        self.widget_id = widget_id or str(uuid.uuid4())
        self._neighbors = [None, None, None, None]

        self.toggle_buttons = toggle_buttons
        self._widgets = list()
        for idx, button in enumerate(toggle_buttons):
            if not isinstance(button, Button):
                raise ValueError(f"Item {idx} {type(button)!r} "
                                 f"is not a Button")
            if button.behavior != Button.TOGGLE:
                raise ValueError(f"Button {idx} is not a TOGGLE")

            # Insert the group control method into the button
            old_function = button.function
            button.function = self.group_function
            self._widgets.append((button, old_function))

    @property
    def neighbors(self):
        return self._neighbors

    @property
    def sub_widgets(self):
        return self._widgets

    def group_function(self, button, state):
        """ Enforce the group behavior; one button on, all others off.

        Parameters
        ----------
        button : ToggleButton
            The button that is receiving the event
        state : bool
            The state being sent to the button

        Returns
        -------
        Optional['Page']
            Returns whatever the button's function returns.  The disabled
            button results are ignored.
        """
        # If we are enabling this button, disable all other buttons
        for b, f in self._widgets:
            if b.id != button.id:
                b._active = False
                if f:
                    f(b, False)

        # Now the actual button press
        f = next(f for b, f in self._widgets if b.id == button.id)
        if f:
            return f(button, state)
        return None

    @property
    def id(self):
        return self.widget_id

    def event(self, event, touch) -> Tuple[bool, Optional['Page']]:
        for button, _ in self._widgets:
            consumed, new_page = button.event(event, touch)
            if consumed:
                return True, new_page
        return False, None

    def focus(self, enabled):
        return

    def can_focus(self) -> bool:
        return False

    def render(self, surface):
        for button, _ in self._widgets:
            button.render(surface)

    def render_focus(self, surface):
        pass



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - An entry page with pulsing lights
"""

# ---- Pulsing light test function (light)

from colors import RED
from page import Page
from tests._utilities import pulsing_light_func
from widget_buttons import Button
from widget_displays import Light, Label

_entry_label = Label((0, 0), (800, 50), label="ENTER")

_pulse1_light = Light((0, 0), (50, 50), function=pulsing_light_func)
_pulse1_light._rate = 400

_pulse2_light = Light((750, 0), (50, 50), function=pulsing_light_func)
_pulse2_light._rate = 300

_pulse3_light = Light((0, 430), (50, 50), function=pulsing_light_func)
_pulse3_light._rate = 200

_pulse4_light = Light((750, 430), (50, 50), pulsing_light_func)
_pulse4_light._rate = 100

def go_to_page_sampler(button, state):
    from tests.test_pages.pages import get_pages
    from tests._utilities  import go_to_page
    return go_to_page(get_pages(), 'sampler')

_touch_me_button = Button((200, 150), (400, 100), font_size=100,
                          label_color1=RED, label="Touch Me",
                          function=go_to_page_sampler)


def entry_page():
    # Tests dynamic Light functions, Label, and default Page function
    _entry_page = Page('entry', [
        _entry_label,
        _pulse1_light,
        _pulse2_light,
        _pulse3_light,
        _pulse4_light,
        _touch_me_button
    ])
    return _entry_page



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Widget Base Classes, providing the core functionality of the widgets.

    Widgets are single interactive elements attached to a Page.
"""
import math
import uuid
from abc import ABC, abstractmethod
from typing import Tuple, Optional

import pygame
# I'm hoping you can init multiple times
pygame.init()

from custom_inherit import doc_inherit
from pygame import Color

from pi_touch_gui._utilities import snapped_value_string
from pi_touch_gui.colors import WHITE, BLACK, MEDIUM_GRAY
from pi_touch_gui.multitouch.raspberrypi_ts import (TS_PRESS, TS_RELEASE,
                                                    TS_MOVE)

# Default font name - specify one here if you want something other
# than the system default.
FONT_NAME = None


class IWidgetInterface(ABC):
    """ The bare minimum methods to implement to be a widget.
    """

    @property
    @abstractmethod
    def id(self):
        """ The Widget ID, uniquely identifies each widget.
        """

    @property
    def sub_widgets(self):
        return None

    @abstractmethod
    def event(self, event, touch) -> Tuple[bool, Optional['Page']]:
        """ Handle touch events, translating them to individual `on_*` calls.

        Parameters
        ----------
        event : int
            Event code defined for the interface; e.g. _widget_bases.py imports
            TS_PRESS, TS_RELEASE, and TS_MOVE
        touch : Touch
            The Touch representing the event, where `touch.position` is the
            (x, y) coordinate of the touch.

        Returns
        -------
        Optional[Page]
            The page to transition to, or None to stay on this page
        """

    @abstractmethod
    def focus(self, enabled):
        """ Set the widget's focus, for the gestural interface.

        Parameters
        ----------
        enabled : bool
            True if this Widget is receiving focus.
        """

    @abstractmethod
    def can_focus(self):
        """ Return True if this widget can accept focus.
        """

    @abstractmethod
    def render(self, surface):
        """ Render to the device pygame Surface.

        Parameters
        ----------
        surface : pygame.Surface
        """

    @abstractmethod
    def render_focus(self, surface):
        """ Render a default focus marker that is the extent.

        Only called if can_focus() returns True, but should be defined anyway.

        Parameters
        ----------
        surface : pygame.Surface
        """


class IWidget(IWidgetInterface, ABC):
    """ Widget base class that processes events and holds common state.

        Sub-classes can define any or all of the three event handlers:

            _on_press(Touch) -> None
            _on_release(Touch) -> Optional[Page]
            _on_move(Touch) -> None

        where:
            `Page` : If not None, jump to that page on the next frame.
    """
    # Cache fonts, no need for each widget to have their own copy
    _font_cache = dict()

    def __init__(self, position, size,
                 widget_id=None,
                 neighbors=None,
                 font_size=None,
                 label=None,
                 color1=None,
                 color2=None,
                 label_color1=None,
                 label_color2=None):
        """ Initialize the widget.

        Most parameters can be ignored and will pick up sensible defaults.

        Parameters
        ----------
        position : Tuple[int, int]
            The (x, y) top-left corner of the widget. The screen positioning
            is accounted at (0, 0) being the far top-left, increasing down
            and to the right.
        size : Tuple[int, int]
            The (w, h) of the widget's rectangular extent.
            Note that in some widgets labels or indicators could fall outside
            of these extents, and other widgets may choose to re-define the
            extent to something less rectangular.
        widget_id : str
            Optional ID for this widget, if not given, gets a random uuid.
        neighbors : List[Union[str, IWidget]]
            The neighbors of this widget, specified by the widget itself or
            that widget's ID.  The list order is [GO_LEFT, GO_RIGHT,
            GO_UP, GO_DOWN].  If you want to stop at an edge, specify
            the neighbor as Page.WALL. These neighbor hints help with
            keyboard arrow navigation of the Page.
        font_size : int
            Height of the font in pixels.  If not specified, will be 1/2 the
            height of the widget.
        label : str
        color1 : Color
            Main color of the widget as a tuple of (r, g, b) or (r, g, b, a),
            or a pygame.Color; the unselected color or foreground color
        color2 : Color
            Secondary color of the widget; selected color or background color
        label_color1 : Union[Tuple[int, int, int], Tuple[int, int, int, int]]
            Main color of the label; unselected color
        label_color2 : Color
            Secondary color of the label; selected color
        """
        self.widget_id = widget_id or str(uuid.uuid4())

        self.position = position
        self._x, self._y = position

        self.size = size
        self._w, self._h = size

        # Fill in any important values not specified
        self.label = label or ""
        self.color1 = color1 or MEDIUM_GRAY
        self.color2 = color2 or WHITE
        self.label_color1 = label_color1 or WHITE
        self.label_color2 = label_color2 or BLACK

        self.font_size = font_size or int(min(self._w, self._h) >> 1)

        font_key = f"{FONT_NAME}:{self.font_size}"
        if font_key in self._font_cache:
            font = self._font_cache[font_key]
        else:
            font = pygame.font.Font(FONT_NAME, self.font_size)
            self._font_cache[font_key] = font
        self._font = font

        self._pressed = False
        self._focus = False
        # We can't currently serialize neighbors, as it causes infinite loops
        # Need to develop a non-object mechanism to name neighbors
        self._neighbors = neighbors

        self._touches = []

        # Probe for the existence of the event handlers.
        try:
            callable(self._on_press)
        except AttributeError:
            self._on_press = None
        try:
            callable(self._on_release)
        except AttributeError:
            self._on_release = None
        try:
            callable(self._on_move)
        except AttributeError:
            self._on_move = None
        try:
            callable(self._adjust)
        except AttributeError:
            self._adjust = None

    @property
    def id(self):
        return self.widget_id

    @property
    def pressed(self) -> bool:
        """ True if the widget is currently pressed.
        """
        return len(self._touches) > 0 or self._pressed

    def set_neighbors(self, new_neighbors):
        """ For enhanced keyboard navigation, point to the neighboring widgets.

        If neighbors are not specified, a more limited keyboard navigation is
        available based on the list order of the widgets in the Page.

        Note that you specify the neighbors as ID or IWidget objects, but on
        use the system takes the liberty of changing the representation to
        whatever it sees fit so don't be surprised if the internal values
        change.

        Parameters
        ----------
        new_neighbors : List(Union(String, IWidget))
            The list of neighbors in order Up, Down, Left, Right; where each
            entry can be an IWidget, the widget_id, or Page.WALL
        """
        if self._neighbors is None:
            self._neighbors = [None, None, None, None]

        for idx, neighbor in enumerate(new_neighbors):
            if isinstance(neighbor, IWidget):
                self._neighbors[idx] = neighbor
            elif isinstance(neighbor, str):
                self._neighbors[idx] = neighbor
            else:
                raise ValueError(f"Unsupported neighbor {idx} type "
                                 f"{type(neighbor)!r}")

    def focus(self, enabled):
        self._focus = enabled

    def can_focus(self) -> bool:
        return True

    def touch_inside(self, touch) -> bool:
        """ Test a touch against the generic widget rectangle.

        Parameters
        ----------
        touch : Touch
            The touch coordinates, where `touch.position` is a Tuple[int, int]
            of the (x, y) position of the touch.

        Returns
        -------
        bool
            True if the touch is inside the widget extents.
        """
        x, y = touch.position
        return (self._x <= x <= self._x + self._w and
                self._y <= y <= self._y + self._h)

    def event(self, event, touch) -> Tuple[bool, Optional['Page']]:
        # PRESS
        # A press can only be registered when it is in the widget bounds
        if self.touch_inside(touch):
            if event == TS_PRESS and touch not in self._touches:
                self._touches.append(touch)
                if self._on_press:
                    return True, self._on_press(touch)

        # RELEASE
        # A touch can be released even when its not over a widget.
        if event == TS_RELEASE and touch in self._touches:
            self._touches.remove(touch)
            if self._on_release:
                return True, self._on_release(touch)

        # MOVE
        # Touch movement is tracked even when it's not over a widget
        if event == TS_MOVE and touch in self._touches:
            if self._on_move:
                return True, self._on_move(touch)

        # "Not Me" - report not consumed
        return False, None

    def press(self):
        self._pressed = True

    def release(self):
        self._pressed = False

    def state_colors(self, active=None) -> Tuple[Color, Color]:
        """ Return the (color, label_color) appropriate to the current state.

        Parameters
        ----------
        active : bool
            Override to force the widget to render as active.

        Returns
        -------
        Tuple[Color, Color]
            Returns a tuple of (color, label_color) where the colors can
            be a pygame.Color or (r,g,b) or (r,g,b,a) tuples.
        """
        active = self.pressed or active is True
        if active:
            color = self.color2
            label_color = self.label_color2
        else:
            color = self.color1
            label_color = self.label_color1
        return color, label_color

    def focus_color(self):
        """ Return a changing color for the focus highlight
        """
        time = pygame.time.get_ticks()
        pulse = math.sin(time / 133) + 1.0  # sets to range 0..2
        return (pulse * 64 + 127,
                64 - pulse * 16,
                64 - pulse * 16)

    def render_centered_text(self, surface, text, color, bg_color=None):
        """ Render text in the center of the widget.

        Parameters
        ----------
        surface : pygame.Surface
        text : str
        color : Color
        bg_color : Color
            If you specify the background color, it can render the text a
            bit faster because it won't use alpha for a clear background.
        """
        if text is None:
            return

        text = self._font.render(text, 1, color, bg_color)
        textpos = text.get_rect()
        textpos.centerx = self._x + (self._w / 2)
        textpos.centery = self._y + (self._h / 2)
        surface.blit(text, textpos)

    def render_left_text(self, surface, text, color, bg_color=None):
        """ Render text in the left side of the widget.

        Parameters
        ----------
        surface : pygame.Surface
        text : str
        color : Color
        bg_color : Color
            If you specify the background color, it can render the text a
            bit faster because it won't use alpha for a clear background.
        """
        if text is None:
            return

        text = self._font.render(text, 1, color, bg_color)
        textpos = text.get_rect()
        textpos.left = self._x
        textpos.centery = self._y + (self._h / 2)
        surface.blit(text, textpos)

    def render_right_text(self, surface, text, color, bg_color=None):
        """ Render text in the right side of the widget.

        Parameters
        ----------
        surface : pygame.Surface
        text : str
        color : Color
        bg_color : Color
            If you specify the background color, it can render the text a
            bit faster because it won't use alpha for a clear background.
        """
        if text is None:
            return

        text = self._font.render(text, 1, color, bg_color)
        textpos = text.get_rect()
        textpos.right = self._x + self._w - 1
        textpos.centery = self._y + (self._h / 2)
        surface.blit(text, textpos)

    def render_focus(self, surface):
        """ Render a default focus outline.

        Parameters
        ----------
        surface : pygame.Surface
        """
        fx, fy = self._x - 2, self._y - 2
        fw, fh = self._w + 4, self._h + 4
        pygame.draw.rect(surface, self.focus_color(), ((fx, fy), (fw, fh)), 4)


class IButtonWidget(IWidget):
    """ ButtonWidget base that extends IWidget for push buttons.
    """

    @doc_inherit(IWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size, function=None, **kwargs):
        """
        Parameters
        ----------
        function : Callable[[IWidget, bool], Optional['Page']]
            Function to call on button events, with a True/False parameter
            that is True when pressed and False when released.
            Optionally returns a new Page to change pages.
        """
        self.function = function
        super(IButtonWidget, self).__init__(position, size, **kwargs)

    def _on_release(self, touch) -> Optional['Page']:
        """ Buttons always call their function on the release event.

        It has already been determined that the touch is inside the
        widget via `touch_inside()`.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for buttons

        Returns
        -------
        Page
            If specified, the new Page to go to.
        """
        return self.release()

    def release(self):
        """ Trigger the button, calling its function if specified.
        """
        super(IButtonWidget, self).release()
        if self.function:
            return self.function(self, False)

    @doc_inherit(IWidget.render, style='numpy_with_merge')
    def render(self, surface):
        """ Render the generic button as a filled rectangle with a label.
        """
        color, label_color = self.state_colors()

        pygame.draw.rect(surface, color, (self.position, self.size), 0)
        self.render_centered_text(surface, self.label, label_color, color)

        if self._focus:
            self.render_focus(surface)


class IControlWidget(IWidget, ABC):
    """ ControlWidget base that extends IWidget for input controls.
    """

    @doc_inherit(IWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size,
                 min_max,
                 start_value=None,
                 snap_value=None,
                 dynamic=None,
                 function=None,
                 **kwargs):
        """
        Parameters
        ----------
        min_max : Tuple[float, float]
            The minimum and maximum values the control might take.
        start_value : float
            The initial value the widget holds.
        snap_value : float
            A value that defines (by example) the display (snap) resolution
            of the widget's value.  E.g. 0.01, 0.1, 1.  The snap value is
            not just a display affect, but controls the resolution of the
            adjusted value of the widget itself.
        dynamic : bool
            True if the function is called each frame the control is active,
            otherwise the function is only called on release.
        function : Callable[[IWidget, float], Optional['Page']]
            Function to call on control events, with a new value for the
            control. Optionally returns a new Page to change pages on release.
        """
        self.function = function
        super(IControlWidget, self).__init__(position, size,
                                             **kwargs)

        # Constrain the snap values to usable range
        self.snap_value = snap_value or 1
        self.snap_value = 0.01 if self.snap_value < 0.01 else self.snap_value

        self.min_max = min_max
        self._min_val, self._max_val = min_max

        # Value is in 0..1, where adjusted_value is in min_val..max_val
        self._range = self._max_val - self._min_val

        if start_value is not None:
            self._adjusted_value = start_value
            self._value = self.normalize_adjusted_value(start_value)
        else:
            self._adjusted_value = min_max[0]
            self._value = self.normalize_adjusted_value(min_max[0])

        self.dynamic = dynamic or False

    def map_value_to_adjusted_value(self, value) -> float:
        """ Map a value in 0..1 to an interpolated value in min_val..max_val.

        Parameters
        ----------
        value : float
            Widget's internal value in 0..1

        Returns
        -------
            Widget's external value in min_val..max_val, snapped to snap_val
        """
        # map [0, 1] to [min_val, max_val] with snap
        interp_value = float(self._min_val + (self._range * value))
        return int(interp_value / self.snap_value + 0.5) * self.snap_value

    def normalize_adjusted_value(self, adjusted_value) -> float:
        """ Normalize a value in min_val..max_val to 0..1

        Parameters
        ----------
        adjusted_value : float
            Widget's external value in min_val..max_val

        Returns
        -------
        float
            Widget's internal value in 0..1
        """
        # map [min_val, max_val] to [0, 1]
        return (adjusted_value - self._min_val) / self._range

    def value_str(self) -> str:
        """ The display version of the widget's external value.

        Returns
        -------
        str
            The string version of self.adjusted_value, which is the value
            the widget represents to the outside world.
        """
        return snapped_value_string(self._adjusted_value, self.snap_value)

    def _on_move(self, touch):
        """ Controls MAY call their function on the move event.

        If the widget was determined to be dynamic, the function is called
        on each move event with a value update.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for controls
        """
        if self.dynamic is True and self.function is not None:
            return self.function(self, self._adjusted_value)

    def _adjust(self, direction):
        """ Adjust the control up or down based on sign(direction)
        """
        step = math.copysign(self.snap_value, direction)
        value = self._adjusted_value + step
        if value < self._min_val:
            value = self._min_val
        elif value > self._max_val:
            value = self._max_val
        self._adjusted_value = value
        self._value = self.normalize_adjusted_value(value)

        if self.function is not None:
            return self.function(self, self._adjusted_value)

    def _on_release(self, touch):
        """ Controls call their function on the release event.

        It has already been determined that the touch is inside the
        widget via `touch_inside()`.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for controls
        """
        self.release()

    def release(self):
        """ Trigger the button, calling its function if specified with
        the adjusted value.
        """
        if self.function is not None:
            return self.function(self, self._adjusted_value)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Utility functions that don't fit in any particular class or module.
"""
import logging
import traceback
from datetime import datetime
from pathlib import Path

LOG = logging.getLogger(__name__)

BACKLOG = logging.getLogger('_backlog')
handler = logging.StreamHandler()
formatter = logging.Formatter("[%(asctime)s] [%(levelname)s] %(message)s")
handler.setFormatter(formatter)
BACKLOG.addHandler(handler)
BACKLOG.propagate = False


def snapped_value_string(value, snap) -> str:
    """ Round (snap) a value for clean display, against a given resolution.

    Parameters
    ----------
    value : float
        Value to snap
    snap : float
        An example of the resolution to snap to.  Current supports .001, .1,
        and "other" (e.g. 1)

    Returns
    -------
    str
        Formatted string of the value, at the snap resolution.
    """
    if snap < 0.1:
        val_str = f"{value:0.2f}"
    elif snap < 1:
        val_str = f"{value:0.1f}"
    else:
        val_str = f"{int(value)}"
    return val_str


def clock_indicator_function():
    """ An indicator function to provide a clock.
    """
    return False, datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def backlog_error(exc, message):
    frame = traceback.extract_tb(exc.__traceback__)[-1]
    BACKLOG.error(f"[{Path(frame.filename).stem}.{frame.lineno}] "
                  f"{message}: {type(exc).__name__}: {exc}")



from pi_touch_gui.colors import *
from pi_touch_gui.gui import GUI
from pi_touch_gui.page import Page
from pi_touch_gui.widget_buttons import *
from pi_touch_gui.widget_controls import *
from pi_touch_gui.widget_displays import *



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Displays are decorative or informative widgets that are not interactive,
    but provide feedback state via a function.
"""
import logging

import pygame
from custom_inherit import doc_inherit

from pi_touch_gui._utilities import snapped_value_string
from pi_touch_gui._widget_bases import IWidget, FONT_NAME

LOG = logging.getLogger(__name__)


class Label(IWidget):
    """ Text Label with a fixed color and no state function.
    """
    # Label alignment
    LEFT = 0
    CENTER = 1
    RIGHT = 2

    def __init__(self, position, size,
                 font_size=None,
                 label=None,
                 label_color1=None,
                 label_alignment=None,
                 **kwargs):
        """ Initialize the label with a subset of the Widget parameters.

        Parameters
        ----------
        position : Tuple[int, int]
            The (x, y) top-left corner of the Label. The screen positioning
            is accounted at (0, 0) being the far top-left, increasing down
            and to the right.
        size : Tuple[int, int]
            The (w, h) of the Label's rectangular extent.
        font_size : int
            Height of the font in pixels.  If not specified, defaults to
            the Label height (size[1])
        label : str
        label_color1 : Union[Tuple[int, int, int], Tuple[int, int, int, int]]
            Color of the Label
        label_alignment : int
            Alignment of the label in the box, one of LEFT, CENTER, or RIGHT
        """
        font_size = font_size or size[1]
        self.alignment = label_alignment or Label.CENTER

        super(Label, self).__init__(position, size,
                                    font_size=font_size,
                                    label=label,
                                    label_color1=label_color1,
                                    label_color2=label_color1)

    def render(self, surface):
        color, label_color = self.state_colors()

        if self.alignment == Label.LEFT:
            self.render_left_text(surface, self.label, label_color)
        elif self.alignment == Label.CENTER:
            self.render_centered_text(surface, self.label, label_color)
        elif self.alignment == Label.RIGHT:
            self.render_right_text(surface, self.label, label_color)

    def can_focus(self):
        return False


class Light(IWidget):
    """ Lights are circles whose color state is determined by their function.
    """

    def __init__(self, position, size, function, **kwargs):
        """ Initialize the light with a subset of the Widget parameters.

        Parameters
        ----------
        position : Tuple[int, int]
            The (x, y) top-left corner of the light. The screen positioning
            is accounted at (0, 0) being the far top-left, increasing down
            and to the right.
        size : Tuple[int, int]
            The (w, h) of the Light's rectangular extent.  The diameter of
            the circle is given by the minimum of w,y.
        function : Callable[[Light], [Color, Color]]
            The function returns the (ring, fill) colors
        """
        # The parent init sets convenience properties like self.x, etc
        super(Light, self).__init__(position, size)

        self.function = function
        self._radius = min(size[0], size[1]) >> 1
        self._center = (self._x + self._w // 2), (self._y + self._h // 2)

    def render(self, surface):
        if self.function is None:
            color1 = self.color1
            color2 = self.color2
        else:
            color1, color2 = self.function(self)
        # Fill
        pygame.draw.circle(surface, color2, self._center, self._radius, 0)
        # Outline ring
        pygame.draw.circle(surface, color1, self._center, self._radius, 2)

    def can_focus(self):
        return False


class Indicator(IWidget):
    """ Indicators are dynamic Labels, controlled by their function.
    """

    @doc_inherit(IWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size,
                 function=None,
                 **kwargs):
        """ Initialize the Indicator.

        If color1 is None, the indicator is colored like a label, with only
        the label colors being used.

        If color1 is specified, then the indicator is set in a filled
        rectangle with the active/inactive color.

        Parameters
        ----------
        function : Callable[[Indicator], [bool, str]]
            The function that defines the content and state of the indicator.
            The bool is an 'active' flag, which sets the active color
            palette if True.  The string sets the indicator label.  If it is
            None, then the indicator defaults to self.label.
        """
        # Set up the boxing behavior and font size.
        kw_font_size = kwargs.pop('font_size', None)
        if kwargs.get('color1') is None:
            self._box = False
            font_size = kw_font_size or size[1]
        else:
            self._box = True
            font_size = kw_font_size or (size[1] >> 1)
        self.function = function

        super(Indicator, self).__init__(position, size,
                                        font_size=font_size,
                                        **kwargs)

    def render(self, surface):
        if self.function:
            active, label = self.function(self)
        else:
            active, label = False, None
        # Default to the internal label
        label = label or self.label

        color, label_color = self.state_colors(active=active)

        if self._box:
            pygame.draw.rect(surface, color, (self.position, self.size), 0)

        self.render_centered_text(surface, label, label_color)

    def can_focus(self):
        return False


class Graph(IWidget):
    """ A dynamic graph.

        Dynamic content is generated by the `function`
    """
    LABEL_INSET = 5

    @doc_inherit(IWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size,
                 min_max=None,
                 quiet=None,
                 function=None,
                 **kwargs):
        """ Initialize the graph button.

        Parameters
        ----------
        min_max : Tuple[Number, Number]
            The minimum and maximum values to scale into the graph.  If not
            specified, will calculate based on the visible data.
        quiet : Bool
            If set and True, quiets the graph (doesn't print the values)
        function : Callable[[IWidget, int], List[dict]]
            The update is called with the number of samples being requested.
            It should return `data` with that many samples or fewer.  If it
            returns fewer than 2 samples, that trace may not be rendered.
            The dictionary in the list defines the color of the trace and
            a list of float with the sample data:
                {"color": <Color>, "data": [<float>]}
        """
        super(Graph, self).__init__(position, size, **kwargs)
        self.min_max = min_max
        self.quiet = quiet
        self.function = function
        if self.function is not None:
            # Determine how many samples the update method returns, so we can
            # scale the font appropriately.
            samples = self.function(self, 1)
            ratio_h = max(2.5, len(samples))
            self._graph_font = pygame.font.Font(FONT_NAME,
                                                int(self._h / ratio_h + 0.5))

    def render(self, surface):
        """ Render the graph.

        The graph width determines how many data samples we ask for.
        The graph(s) are then drawn inside of the rectangle, with the last
        values in the graph optinally data rendered as labels to the right
        (but inside) of the graph.
        """
        graph_pos = (self._x, self._y)
        graph_size = (self._w, self._h)
        pygame.draw.rect(surface, self.color1, (graph_pos, graph_size), 2)

        # We can only draw the graph(s) if there is an update method
        if self.function is None:
            return

        values = self.function(self, graph_size[0] - 2)
        if self.min_max is None:
            min_val = 9999
            max_val = -9999
            # Determine the global min/max to scale the vertical range
            for entry in values:
                data = entry["data"]
                if len(data) > 0:
                    min_val = entry.get("min", min(min_val, min(data)))
                    max_val = entry.get("min", max(max_val, max(data)))
                else:
                    min_val = 0
                    max_val = 0
        else:
            min_val = self.min_max[0]
            max_val = self.min_max[1]


        range = max_val - min_val
        # Range must be > 0 so adjust around the existing min/max
        if range < .01:
            min_val -= .01
            max_val += .01
            range = max_val - min_val

        for entry in values:
            color = entry["color"]
            data = entry["data"]

            def scale_y_value(value):
                """ Scale the value to fit the graph box."""
                inset_range = range * 1.1
                center_val = value - (range / 2.0)
                return self._y + (self._h / 2.0
                                  - ((center_val - min_val) / inset_range)
                                  * self._h)

            # Need two data points to start the graph
            if len(data) > 1:
                prev_y = scale_y_value(data[0])
                prev_x = graph_pos[0]
                for val in data:
                    x = prev_x + 1
                    y = scale_y_value(val)
                    pygame.draw.line(surface, color, (prev_x, prev_y), (x, y),
                                     2)
                    prev_x = x
                    prev_y = y

                # Render the graph trailing value
                if self.quiet is not True:
                    text = self._graph_font.render(
                        snapped_value_string(data[-1], 0.1),
                        1,
                        color)
                    textpos = text.get_rect()
                    textpos.right = (self._x + self._w) - self.LABEL_INSET
                    # The text tracks the graph line, unless it's outside the
                    # extents, then we clamp it to inside.
                    if (prev_y + textpos.h + self.LABEL_INSET) > (
                            self._y + self._h):
                        textpos.bottom = prev_y - self.LABEL_INSET
                    elif (prev_y - textpos.h - self.LABEL_INSET) < self._y:
                        textpos.top = prev_y + self.LABEL_INSET
                    else:
                        textpos.centery = prev_y
                    surface.blit(text, textpos)

        if self._focus:
            self.render_focus(surface)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - Widget Sample, all the widgets, configured for keyboard grid.
"""

import pygame

from colors import (WHITE, MEDIUM_GRAY, BLUE, RED, BLACK, DARK_RED, GREEN,
                    CYAN, MAGENTA)
from page import Page
from tests._utilities import fake_data_graph_function
from widget_buttons import Button, ButtonGroup
from widget_controls import Dial, Slider
from widget_displays import Indicator, Label, Graph

_boop_count = 4

_up_down_flag = 0

_max_track_samples = 1000
_last_track_update = pygame.time.get_ticks()
_track_1_data = []
_track_2_data = []
_track_3_data = []
_dial_value = 50
_slider_h_value = 40
_slider_v_value = 60


# =========================================================
# Hold-behavior buttons and indicator
# =========================================================

def _up_dn_button_func(button, pushed):
    global _up_down_flag
    direction = button._direction
    if pushed:
        _up_down_flag = direction
    else:
        _up_down_flag = 0


def _up_dn_indicator_func(indicator):
    global _up_down_flag
    if _up_down_flag == 0:
        return False, ''
    elif _up_down_flag == 1:
        return True, "UP"
    else:
        return True, "DN"


_up_button = Button((115, 105), (60, 50),
                    label="^",
                    behavior=Button.HOLD,
                    style=Button.OBROUND,
                    function=_up_dn_button_func)
_up_button._direction = 1

_up_dn_indicator = Indicator((115, 160), (60, 50),
                             label_color2=WHITE,
                             function=_up_dn_indicator_func)

_dn_button = Button((115, 215), (60, 50),
                    label="v",
                    behavior=Button.HOLD,
                    style=Button.OBROUND,
                    function=_up_dn_button_func)
_dn_button._direction = -1


# =========================================================
# Rounded Button styles
# =========================================================
def go_to_page_entry(button, state):
    from tests.test_pages.pages import get_pages
    from tests._utilities import go_to_page
    return go_to_page(get_pages(), 'entry')


_back_button = Button((110, 400), (75, 50),
                      label="Back",
                      style=Button.LEFT_ROUND,
                      function=go_to_page_entry)

_do_nothing_button = Button((190, 400), (150, 50),
                            label="Do Nothing",
                            style=Button.RECTANGLE,
                            color1=MEDIUM_GRAY, color2=WHITE,
                            label_color1=BLUE, label_color2=RED)


def go_to_page_poweroff(button, state):
    from tests.test_pages.pages import get_pages
    from tests._utilities import go_to_page
    return go_to_page(get_pages(), 'power_off')


_next_button = Button((345, 400), (75, 50),
                      label="Next",
                      style=Button.RIGHT_ROUND,
                      function=go_to_page_poweroff)


# ---- Boop button, for some amusement

def _boop_button_func(button, pushed):
    global _boop_count
    _boop_count -= 1
    return None


def _boop_indicator_func(indicator):
    """ Countdown; when we reach 0, we exit the program.
    """
    if _boop_count == 4:
        return False, ''
    if _boop_count == 0:
        exit(0)
    return True, f"{_boop_count}"


_boop_button = Button((430, 400), (50, 50),
                      label="X",
                      style=Button.OBROUND,
                      function=_boop_button_func)

_boop_indicator = Indicator((485, 400), (50, 50),
                            color1=BLACK, color2=RED,
                            label_color1=WHITE, label_color2=BLACK,
                            function=_boop_indicator_func)


# =========================================================
# Toggle-behavior Buttons with reset
# =========================================================

def _reset_count_buttons(button, pressed):
    _one_button.clear()
    _two_button.clear()
    _three_button.clear()


_count_buttons_label = Label((190, 105), (100, 25),
                             label="Pick Several")
_one_button = Button((190, 130), (100, 50),
                     label="One",
                     behavior=Button.TOGGLE)
_two_button = Button((190, 185), (100, 50),
                     label="Two",
                     behavior=Button.TOGGLE)
_three_button = Button((190, 240), (100, 50),
                       label="Three",
                       behavior=Button.TOGGLE)
_count_reset_button = Button((190, 295), (100, 50),
                             label='RESET',
                             label_color1=DARK_RED,
                             function=_reset_count_buttons)

# =========================================================
# Button Group
# =========================================================

_color_buttons_label = Label((300, 105), (100, 25),
                             label="Pick One")

_red_button = Button((300, 130), (100, 50),
                     label="Red",
                     color2=RED,
                     behavior=Button.TOGGLE)
_green_button = Button((300, 185), (100, 50),
                       label="Green",
                       color2=GREEN,
                       behavior=Button.TOGGLE)
_blue_button = Button((300, 240), (100, 50), label="Blue",
                      color2=BLUE,
                      behavior=Button.TOGGLE)

_color_group = ButtonGroup([_red_button, _green_button, _blue_button])


# =========================================================
# Controls and Graphs
# =========================================================

# ---- Graph and Control Functions

def _dial_control_func(dial, value):
    global _dial_value
    _dial_value = value


def _slider_h_control_func(slider, value):
    global _slider_h_value
    _slider_h_value = value


def _slider_v_control_func(slider, value):
    global _slider_v_value
    _slider_v_value = value


def _graph_button_update(graph, size):
    global _dial_value, _slider_h_value, _slider_v_value
    global _max_track_samples
    global _track_1_data, _track_2_data, _track_3_data
    global _last_track_update

    # Update the graph
    time = pygame.time.get_ticks()
    # Update twice per second, to show some life
    if (time - _last_track_update) > 250:
        _track_1_data.append(_dial_value)
        _track_2_data.append(_slider_h_value)
        _track_3_data.append(_slider_v_value)
        _last_track_update = time

    # Truncate to the maximum size to prevent memory crash
    if len(_track_1_data) > _max_track_samples:
        _track_1_data = _track_1_data[-_max_track_samples:]
        _track_2_data = _track_2_data[-_max_track_samples:]
        _track_3_data = _track_3_data[-_max_track_samples:]

    # Return what was asked for
    return [{"color": CYAN, "data": _track_1_data[-size:]},
            {"color": BLUE, "data": _track_2_data[-size:]},
            {"color": MAGENTA, "data": _track_3_data[-size:]}, ]


# ---- Graph and Control Widgets

_control_graph = Graph((410, 130), (390, 100),
                       min_max=(0, 100), function=_graph_button_update)
_control_dial = Dial((450, 240), (150, 150),
                     min_max=(0, 100), start_value=_dial_value,
                     snap_value=1,
                     color2=CYAN,
                     dynamic=True, function=_dial_control_func)
_h_slider = Slider((620, 400), (150, 50),
                   min_max=(25, 75), start_value=_slider_h_value,
                   snap_value=5,
                   label_color1=BLACK, color1=MEDIUM_GRAY, color2=MAGENTA,
                   dynamic=True, function=_slider_v_control_func)
_v_slider = Slider((620, 240), (50, 150),
                   min_max=(0, 100), start_value=_slider_v_value,
                   snap_value=.1,
                   label_color1=BLACK, color1=MEDIUM_GRAY, color2=BLUE,
                   dynamic=False, function=_slider_h_control_func)
_fake_graph = Graph((680, 240), (110, 150), min_max=(-100, 100),
                    quiet=True, function=fake_data_graph_function)

# =========================================================
# Define keyboard navigation grid
# This could probably be automated someday, or bake in
# some kind of layout manager
# =========================================================

_up_button._neighbors = [Page.WALL, _dn_button, Page.WALL, _one_button]
_dn_button._neighbors = [_up_button, _back_button, Page.WALL, _three_button]

_back_button._neighbors = [_dn_button, Page.WALL,
                           Page.WALL, _do_nothing_button]
_do_nothing_button._neighbors = [_count_reset_button, Page.WALL,
                                 _back_button, _next_button]

_next_button._neighbors = [_blue_button, Page.WALL,
                           _do_nothing_button, _boop_button]

_boop_button._neighbors = [_control_dial, Page.WALL,
                           _next_button, _h_slider]

_one_button._neighbors = [Page.WALL, _two_button,
                          _up_button, _red_button]
_two_button._neighbors = [_one_button, _three_button,
                          _up_button, _green_button]
_three_button._neighbors = [_two_button, _count_reset_button,
                            _dn_button, _blue_button]
_count_reset_button._neighbors = [_three_button, _do_nothing_button,
                                  _dn_button, _control_dial]

_red_button._neighbors = [Page.WALL, _green_button,
                          _one_button, Page.WALL]
_green_button._neighbors = [_red_button, _blue_button,
                            _two_button, Page.WALL]
_blue_button._neighbors = [_green_button, _next_button,
                           _three_button, _control_dial]

_control_dial._neighbors = [Page.WALL, _boop_button,
                            _count_reset_button, _v_slider]
_v_slider._neighbors = [Page.WALL, _h_slider, _control_dial, Page.WALL]
_h_slider._neighbors = [_v_slider, Page.WALL, _boop_button, Page.WALL]


# =========================================================
# Page configuration and definition
# =========================================================


def sampler_page(background):
    # Tests background image, Button, RoundedButton, DynamicButton,
    # ToggleButton, ButtonGroup, Label, Indicator, ...
    _sampler_page = Page('sampler',
                         background=background,
                         widgets=[
                             # Putting a non-focus element first to test
                             # that case in keyboard control
                             _up_dn_indicator, _up_button, _dn_button,
                             _back_button, _do_nothing_button, _next_button,
                             _boop_button, _boop_indicator,

                             _count_buttons_label,
                             _one_button, _two_button, _three_button,
                             _count_reset_button,

                             _color_buttons_label,
                             _red_button, _green_button, _blue_button,
                             _color_group,

                             _control_graph, _control_dial, _v_slider,
                             _h_slider,
                             _fake_graph
                         ])

    return _sampler_page
